type entry =
    File of string * file_info
  | Directory of string * entry list
  | Error of string * exn
and file_info = { file_size : int; }
val with_chdir_to : string -> (unit -> 'a) -> 'a
val slurp : string -> entry
