(* Log *)

val sf : ('a, unit, string) format -> 'a
val enable : bool ref
val init : unit -> unit
val shutdown : unit -> unit
val set_file : string -> unit
val set_enable : bool -> unit
val log : [`Debug|`Auth|`Access|`Error|`Exception|`Info] -> string -> unit
val debug : string -> unit
val auth : string -> unit
val access : string -> unit
val error : string -> unit
val info : string -> unit
val exc : exn -> string -> unit
