val config_file : string ref
val user_specified_config_file : bool ref
val dump_config : bool ref
val low_memory : int ref
val fast : bool ref
val very_slow : bool ref
val specs : (string * Arg.spec * string) list
