module C :
  sig
    val home : string
    val directory : string
    val primary_file : unit -> string
    val current : Configurator.configurator
    module Convert : sig val convert : Oldconfig.t -> Configfile.t end
    val load : unit -> (string * exn) list
  end
val home : string
val directory : string
val primary_file : unit -> string
val current : Configurator.configurator
module Convert : sig val convert : Oldconfig.t -> Configfile.t end
val load : unit -> (string * exn) list
