val sf : ('a, unit, string) format -> 'a
val first_line : string -> string
val limit : int -> string -> string
val for_all_chars : (char -> bool) -> string -> bool
val split_once_at : (char -> bool) -> string -> string * string
val is_digit : char -> bool
val is_space : char -> bool
val parse_strings : string -> string list
val split_at : char -> string -> string list
val list_intersect : 'a list -> 'a list -> 'a list
val once : (unit -> unit) -> unit -> unit
val list_has_more_than_one_element : 'a list -> bool
val count_lines : string -> int
val first_matching_char_from : int -> (char -> bool) -> string -> int
val first_matching_char : (char -> bool) -> string -> int
val longest_matching_prefix : (char -> bool) -> string -> string * string
val remove_leading_spaces : string -> string
val delete_first_chars : int -> string -> string
val hierarchical : string -> string -> int
val wind : ('a -> 'b) -> 'a -> ('c -> 'd) -> 'c -> 'b
val list_change_nth : 'a list -> int -> 'a -> 'a list
val list_remove_nth : 'a list -> int -> 'a list
val word_wrap : out_channel -> ?columns:int -> string -> unit
val reg_of_string : string -> string
val flip_array : 'a array -> unit
val proc_get_rsz_vsz : unit -> int * int
val proc_get_free_mem : unit -> int
val substitute_variables : (string * string) list -> string -> string
val string_of_process_status : string -> Unix.process_status -> string option
val list_sub_rev : 'a list -> int -> int -> 'a list
