/* $Id: default.css,v 1.10 2004/10/31 20:12:15 berke Exp $ */
/* Copyright (C) 2002 Berke Durak. All rights reserved. */
/* This stylesheet is under the GNU General Public License. */
/* This means you can use and modify it freely, but you have */
/* to credit me as your original source of inspiration. */
body { 
  margin-left: 0.2em;
  background-color: #ffffff;
  color: #000000;
  font-family: helvetica, sans-serif;
}

ul {
  list-style-type: square;
}

h2 {
  font-size: 1.5em;
  font-family: helvetica, sans-serif;
  font-weight: bold;
  text-align: left;
  border-bottom-width: 0px;
  border-left-width: 0px;
  border-right-width: 0px;
  border-top-width: 0px;
  padding-left: 0px;
  padding-top: 0px;
  border-color: #000000;
  border-style: solid;
  margin-bottom: 8px;
}

h1 {
  width: 100%;
  border-bottom-width: 8px;
  border-top-width: 0px;
  border-left-width: 0px;
  border-right-width: 0px;
  border-style: solid;
  border-color: #6080f0;
}

div.bottom {
  width: 100%;
  border-top-width: 8px;
  border-bottom-width: 0px;
  border-left-width: 0px;
  border-right-width: 0px;
  border-style: solid;
  border-color: #6080f0;
}

div.signature {
  text-align: left;
  float: left;
  font-style: italic;
} 

div.compliance {
  text-align: right;
  width: 100%;
}

div.query
{ font-family: courier, sans-serif;
  background-color: #ffff00;
  color: #000000; }

div.query span.highlight
{ background-color: #ff0000; }

a {
  text-decoration: underline;
  color: #0000ff;
}

a:visited {
  text-decoration: underline;
  color: #8080ff;
}

a:hover {
  color: #2030ff;
  background-color: #ffff00;
}

div.statistics {
   color: #808080;
   font-size: 0.7em;
   text-align:right;
}
