(***********************************************************************)
(*                                                                     *)
(*                               Ledit                                 *)
(*                                                                     *)
(*       Daniel de Rauglaudre, projet Cristal, INRIA Rocquencourt      *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ledit.mli,v 1.3 1997/10/28 10:01:04 ddr Exp $ *)

val init : unit -> unit
val input_char : in_channel -> char
val read_line : unit -> string
val set_prompt : string -> unit
val get_prompt : unit -> string
val open_histfile : bool -> string -> unit
val close_histfile : unit -> unit
val set_max_len : int -> unit
val set_son : int -> unit
