class counter :
  object
    val mutable col : int
    val mutable row : int
    method col : int
    method private count_char : char -> unit
    method flush : unit
    method output_char : char -> unit
    method output_string : string -> unit
    method output_substring : string -> int -> int -> unit
    method row : int
  end
class writer_of_output_channel :
  out_channel ->
  object
    val mutable col : int
    val mutable row : int
    method col : int
    method private count_char : char -> unit
    method flush : unit
    method output_char : char -> unit
    method output_string : string -> unit
    method output_substring : string -> int -> int -> unit
    method row : int
  end
class word_wrapper :
  ?columns:int ->
  < output_char : char -> unit; output_string : string -> 'a;
    output_substring : string -> int -> int -> 'b; .. > ->
  object
    val mutable j : int
    method columns : int
    method flush : unit
    method newline : unit
    method output : string -> unit
    method output_word : string -> unit
  end
class word_non_wrapper :
  ?columns:int ->
  < col : int; output_char : char -> unit; output_string : string -> unit;
    .. > ->
  object
    method columns : int
    method flush : unit
    method newline : unit
    method output : string -> unit
    method output_word : string -> unit
  end
