exception Parse_error of int * int * string
val lexical_error : Lexing.lexbuf -> string -> 'a
val token : Lexing.lexbuf -> Syntax.token
val eat_comment : Lexing.lexbuf -> unit
val readstr : Lexing.lexbuf -> string
val readregexp : Lexing.lexbuf -> string * Ast.regexp_option list
val readregexpoptions : Lexing.lexbuf -> Ast.regexp_option list
val readident : Lexing.lexbuf -> string
